extern crate nihav;

use std::io::SeekFrom;
use std::fs::File;
use std::path::Path;
use nihav::io::byteio::{FileReader, ByteReader};
use nihav::frame::*;
use nihav::codecs::*;
use nihav::demuxers::*;
use nihav::detect;
use std::env;

mod frmwriter;
use frmwriter::*;

mod wavwriter;
use wavwriter::WavWriter;

struct FrameOutput {
    prefix:     &'static str,
    streamno:   usize,
}

impl FrameOutput {
    fn output_frame(&mut self, pkt: &NAPacket, frmref: NAFrameRef) {
        let frm = frmref.borrow();
        if frm.get_frame_type() != FrameType::Skip {
            let pts = if let Some(pt) = frm.get_pts() { pt } else { pkt.get_pts().unwrap() };
            let frmbuf = frm.get_buffer().get_vbuf().unwrap();
            if frmbuf.get_info().get_format().is_paletted() {
                write_palppm(self.prefix, self.streamno, pts, frm);
            } else {
                write_pgmyuv(self.prefix, self.streamno, pts, frm);
            }
        }
    }
}

struct AudioOutput {
    wwr:    WavWriter<'static>,
    wrote_header: bool,
}

impl AudioOutput {
    fn new(name: &String) -> Self { Self { wwr: WavWriter::new(name), wrote_header: false } }
    fn output_frame(&mut self, _pkt: &NAPacket, frm: NAFrameRef) {
        if !self.wrote_header {
            self.wwr.write_header(frm.borrow().get_info().as_ref().get_properties().get_audio_info().unwrap()).unwrap();
            self.wrote_header = true;
        }
        self.wwr.write_frame(frm.borrow().get_buffer()).unwrap();
    }
}

enum Outputter {
    Video(FrameOutput),
    Audio(AudioOutput),
    None,
}

fn main() {
    let args: Vec<_> = env::args().collect();

    if args.len() == 1 {
        println!("usage: nihav-tool [-noout] [-vn] [-an] input [lastpts]");
        return;
    }
    let mut lastpts: Option<u64> = None;
    let mut cur_arg: usize = 1;
    let mut noout = false;
    let mut decode_video = true;
    let mut decode_audio = true;

    while (cur_arg < args.len()) && (args[cur_arg].chars().next().unwrap() == '-') {
        match args[cur_arg].as_str() {
            "--"        => { break; },
            "-noout"    => { noout = true; },
            "-an"       => { decode_audio = false; },
            "-vn"       => { decode_video = false; },
            _           => { println!("unknown option {}", args[cur_arg]); return; },
        }
        cur_arg += 1;
    }
    let name = args[cur_arg].as_str();
    cur_arg += 1;
    if cur_arg < args.len() {
        lastpts = Some(u64::from_str_radix(args[cur_arg].as_str(), 10).unwrap());
    }

    let path = Path::new(name);
    let mut file = File::open(path).unwrap();
    let dmx_fact;
    let mut fr = FileReader::new_read(&mut file);
    let mut br = ByteReader::new(&mut fr);
    let res = detect::detect_format(name, &mut br);
    if res.is_none() {
        println!("cannot detect format for {}", name);
        return;
    }
    let (dmx_name, _) = res.unwrap();
println!("trying demuxer {} on {}", dmx_name, name);
    dmx_fact = find_demuxer(dmx_name).unwrap();
    br.seek(SeekFrom::Start(0)).unwrap();
    let mut dmx = create_demuxer(dmx_fact, &mut br).unwrap();

    let mut decs: Vec<Option<Box<NADecoder>>> = Vec::new();
    let mut writers: Vec<Outputter> = Vec::new();
    for i in 0..dmx.get_num_streams() {
        let s = dmx.get_stream(i).unwrap();
        let info = s.get_info();
        let decfunc = find_decoder(info.get_name());
println!("stream {} - {} {}", i, s, info.get_name());
        let mut has_out = false;
        if info.is_video() {
            if decode_video {
                let mut dec = (decfunc.unwrap())();
                dec.init(info).unwrap();
                decs.push(Some(dec));
                if !noout {
                    writers.push(Outputter::Video(FrameOutput{prefix: "", streamno: i}));
                    has_out = true;
                }
            } else {
                decs.push(None);
            }
        } else if info.is_audio() {
            if decode_audio {
                let mut dec = (decfunc.unwrap())();
                dec.init(info).unwrap();
                decs.push(Some(dec));
                if !noout {
                    let name = format!("out{:02}.wav", i);
                    writers.push(Outputter::Audio(AudioOutput::new(&name)));
                    has_out = true;
                }
            } else {
                decs.push(None);
            }
        } else {
            decs.push(None);
panic!("decoder {} not found", info.get_name());
        }
        if !has_out {
            writers.push(Outputter::None);
        }
    }

    loop {
        let pktres = dmx.get_frame();
        if let Err(e) = pktres {
            if e == DemuxerError::EOF { break; }
        }
        let pkt = pktres.unwrap();
        let streamno = pkt.get_stream().get_id() as usize;
        if let Some(ref mut dec) = decs[streamno] {
            let frm = dec.decode(&pkt).unwrap();
            if !noout {
                match writers[streamno] {
                    Outputter::Video(ref mut wr) => { wr.output_frame(&pkt, frm); },
                    Outputter::Audio(ref mut wr) => { wr.output_frame(&pkt, frm); },
                    _ => {},
                };
            }
        }
        if pkt.get_pts() != None && pkt.get_pts() >= lastpts { break; }
    }
//panic!("end");
}
